/*
  Copyright (C) 2010-2011  Charles Pence <charles@charlespence.net>
  Copyright (C) 1996-1997  Id Software, Inc.

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "pch.h"
#include "quakedef.h"

/*

The view is allowed to move slightly from it's true position for bobbing,
but if it exceeds 8 pixels linear distance (spherical, not box), the list of
entities sent from the server may not include everything in the pvs, especially
when crossing a water boudnary.

*/

float    v_dmg_time, v_dmg_roll, v_dmg_pitch;

extern    int            in_forward, in_forward2, in_back;


/*
===============
V_CalcRoll

Used by view and sv_user
===============
*/
vec3_t    forward, right, up;

float V_CalcRoll (vec3_t angles, vec3_t velocity)
{
    float    sign;
    float    side;
    float    value;

    AngleVectors (angles, forward, right, up);
    side = DotProduct (velocity, right);
    sign = side < 0 ? -1 : 1;
    side = fabs(side);

    value = Cvar_VariableValue("cl_rollangle");

    if (side < Cvar_VariableValue("cl_rollspeed"))
        side = side * value / Cvar_VariableValue("cl_rollspeed");
    else
        side = value;

    return side*sign;

}


/*
===============
V_CalcBob

===============
*/
float V_CalcBob (void)
{
    float    bob;
    float    cycle;

    cycle = cl.time - (int)(cl.time / Cvar_VariableValue("cl_bobcycle")) * Cvar_VariableValue("cl_bobcycle");
    cycle /= Cvar_VariableValue("cl_bobcycle");
    if (cycle < Cvar_VariableValue("cl_bobup"))
        cycle = M_PI * cycle / Cvar_VariableValue("cl_bobup");
    else
        cycle = M_PI + M_PI * (cycle - Cvar_VariableValue("cl_bobup"))/(1.0 - Cvar_VariableValue("cl_bobup"));

    // bob is proportional to velocity in the xy plane
    // (don't count Z, or jumping messes it up)
    bob = sqrt(cl.velocity[0]*cl.velocity[0] + cl.velocity[1]*cl.velocity[1]) * Cvar_VariableValue("cl_bob");

    bob = bob * 0.3 + bob * 0.7 * sin(cycle);
    if (bob > 4)
        bob = 4;
    else if (bob < -7)
        bob = -7;
    return bob;
}


//=============================================================================

void V_StartPitchDrift (void)
{
#if 1
    if (cl.laststop == cl.time)
    {
        return;        // something else is keeping it from drifting
    }
#endif
    if (cl.nodrift || !cl.pitchvel)
    {
        cl.pitchvel = Cvar_VariableValue("v_centerspeed");
        cl.nodrift = false;
        cl.driftmove = 0;
    }
}

void V_StopPitchDrift (void)
{
    cl.laststop = cl.time;
    cl.nodrift = true;
    cl.pitchvel = 0;
}

/*
===============
V_DriftPitch

Moves the client pitch angle towards cl.idealpitch sent by the server.

If the user is adjusting pitch manually, either with lookup/lookdown,
mlook and mouse, or klook and keyboard, pitch drifting is constantly stopped.

Drifting is enabled when the center view key is hit, mlook is released and
lookspring is non 0, or when
===============
*/
void V_DriftPitch (void)
{
    float        delta, move;

    if (noclip_anglehack || !cl.onground || cls.demoplayback )
    {
        cl.driftmove = 0;
        cl.pitchvel = 0;
        return;
    }

// don't count small mouse motion
    if (cl.nodrift)
    {
        if (fabs(cl.cmd.forwardmove) < Cvar_VariableValue("cl_forwardspeed"))
            cl.driftmove = 0;
        else
            cl.driftmove += host_frametime;

        if (cl.driftmove > Cvar_VariableValue("v_centermove"))
            V_StartPitchDrift();

        return;
    }

    delta = cl.idealpitch - cl.viewangles[PITCH];

    if (!delta)
    {
        cl.pitchvel = 0;
        return;
    }

    move = host_frametime * cl.pitchvel;
    cl.pitchvel += host_frametime * Cvar_VariableValue("v_centerspeed");

    if (delta > 0)
    {
        if (move > delta)
        {
            cl.pitchvel = 0;
            move = delta;
        }
        cl.viewangles[PITCH] += move;
    }
    else if (delta < 0)
    {
        if (move > -delta)
        {
            cl.pitchvel = 0;
            move = -delta;
        }
        cl.viewangles[PITCH] -= move;
    }
}





/*
==============================================================================

                        PALETTE FLASHES

==============================================================================
*/


cshift_t    cshift_empty = { {130,80,50}, 0 };
cshift_t    cshift_water = { {130,80,50}, 128 };
cshift_t    cshift_slime = { {0,25,5}, 150 };
cshift_t    cshift_lava = { {255,80,0}, 150 };

byte        ramps[3][256];
float        v_blend[4];        // rgba 0.0 - 1.0


/*
===============
V_ParseDamage
===============
*/
void V_ParseDamage (void)
{
    int        armor, blood;
    vec3_t    from;
    int        i;
    vec3_t    forward, right, up;
    entity_t    *ent;
    float    side;
    float    count;

    armor = MSG_ReadByte ();
    blood = MSG_ReadByte ();
    for (i=0 ; i<3 ; i++)
        from[i] = MSG_ReadCoord ();

    count = blood*0.5 + armor*0.5;
    if (count < 10)
        count = 10;

    cl.faceanimtime = cl.time + 0.2;        // but sbar face into pain frame

    cl.cshifts[CSHIFT_DAMAGE].percent += 3*count;
    if (cl.cshifts[CSHIFT_DAMAGE].percent < 0)
        cl.cshifts[CSHIFT_DAMAGE].percent = 0;
    if (cl.cshifts[CSHIFT_DAMAGE].percent > 150)
        cl.cshifts[CSHIFT_DAMAGE].percent = 150;

    if (armor > blood)
    {
        cl.cshifts[CSHIFT_DAMAGE].destcolor[0] = 200;
        cl.cshifts[CSHIFT_DAMAGE].destcolor[1] = 100;
        cl.cshifts[CSHIFT_DAMAGE].destcolor[2] = 100;
    }
    else if (armor)
    {
        cl.cshifts[CSHIFT_DAMAGE].destcolor[0] = 220;
        cl.cshifts[CSHIFT_DAMAGE].destcolor[1] = 50;
        cl.cshifts[CSHIFT_DAMAGE].destcolor[2] = 50;
    }
    else
    {
        cl.cshifts[CSHIFT_DAMAGE].destcolor[0] = 255;
        cl.cshifts[CSHIFT_DAMAGE].destcolor[1] = 0;
        cl.cshifts[CSHIFT_DAMAGE].destcolor[2] = 0;
    }

//
// calculate view angle kicks
//
    ent = &cl_entities[cl.viewentity];

    VectorSubtract (from, ent->origin, from);
    VectorNormalize (from);

    AngleVectors (ent->angles, forward, right, up);

    side = DotProduct (from, right);
    v_dmg_roll = count * side * Cvar_VariableValue("v_kickroll");

    side = DotProduct (from, forward);
    v_dmg_pitch = count * side * Cvar_VariableValue("v_kickpitch");

    v_dmg_time = Cvar_VariableValue("v_kicktime");
}


/*
==================
V_cshift_f
==================
*/
void V_cshift_f (void)
{
    cshift_empty.destcolor[0] = Util::lexical_cast<int>(Cmd_Argv(1));
    cshift_empty.destcolor[1] = Util::lexical_cast<int>(Cmd_Argv(2));
    cshift_empty.destcolor[2] = Util::lexical_cast<int>(Cmd_Argv(3));
    cshift_empty.percent = Util::lexical_cast<int>(Cmd_Argv(4));
}


/*
==================
V_BonusFlash_f

When you run over an item, the server sends this command
==================
*/
void V_BonusFlash_f (void)
{
    cl.cshifts[CSHIFT_BONUS].destcolor[0] = 215;
    cl.cshifts[CSHIFT_BONUS].destcolor[1] = 186;
    cl.cshifts[CSHIFT_BONUS].destcolor[2] = 69;
    cl.cshifts[CSHIFT_BONUS].percent = 50;
}

/*
=============
V_SetContentsColor

Underwater, lava, etc each has a color shift
=============
*/
void V_SetContentsColor (int contents)
{
    switch (contents)
    {
    case CONTENTS_EMPTY:
    case CONTENTS_SOLID:
        cl.cshifts[CSHIFT_CONTENTS] = cshift_empty;
        break;
    case CONTENTS_LAVA:
        cl.cshifts[CSHIFT_CONTENTS] = cshift_lava;
        break;
    case CONTENTS_SLIME:
        cl.cshifts[CSHIFT_CONTENTS] = cshift_slime;
        break;
    default:
        cl.cshifts[CSHIFT_CONTENTS] = cshift_water;
    }
}

/*
=============
V_CalcPowerupCshift
=============
*/
void V_CalcPowerupCshift (void)
{
    if (cl.items & IT_QUAD)
    {
        cl.cshifts[CSHIFT_POWERUP].destcolor[0] = 0;
        cl.cshifts[CSHIFT_POWERUP].destcolor[1] = 0;
        cl.cshifts[CSHIFT_POWERUP].destcolor[2] = 255;
        cl.cshifts[CSHIFT_POWERUP].percent = 30;
    }
    else if (cl.items & IT_SUIT)
    {
        cl.cshifts[CSHIFT_POWERUP].destcolor[0] = 0;
        cl.cshifts[CSHIFT_POWERUP].destcolor[1] = 255;
        cl.cshifts[CSHIFT_POWERUP].destcolor[2] = 0;
        cl.cshifts[CSHIFT_POWERUP].percent = 20;
    }
    else if (cl.items & IT_INVISIBILITY)
    {
        cl.cshifts[CSHIFT_POWERUP].destcolor[0] = 100;
        cl.cshifts[CSHIFT_POWERUP].destcolor[1] = 100;
        cl.cshifts[CSHIFT_POWERUP].destcolor[2] = 100;
        cl.cshifts[CSHIFT_POWERUP].percent = 100;
    }
    else if (cl.items & IT_INVULNERABILITY)
    {
        cl.cshifts[CSHIFT_POWERUP].destcolor[0] = 255;
        cl.cshifts[CSHIFT_POWERUP].destcolor[1] = 255;
        cl.cshifts[CSHIFT_POWERUP].destcolor[2] = 0;
        cl.cshifts[CSHIFT_POWERUP].percent = 30;
    }
    else
        cl.cshifts[CSHIFT_POWERUP].percent = 0;
}

/*
=============
V_CalcBlend
=============
*/
void V_CalcBlend (void)
{
    float    r, g, b, a, a2;
    int        j;

    r = 0;
    g = 0;
    b = 0;
    a = 0;

    for (j=0 ; j<NUM_CSHIFTS ; j++)
    {
        if (!Cvar_VariableValue("gl_cshiftpercent"))
            continue;

        a2 = ((cl.cshifts[j].percent * Cvar_VariableValue("gl_cshiftpercent")) / 100.0) / 255.0;
        if (!a2)
            continue;

        a = a + a2*(1-a);
        a2 = a2/a;
        r = r*(1-a2) + cl.cshifts[j].destcolor[0]*a2;
        g = g*(1-a2) + cl.cshifts[j].destcolor[1]*a2;
        b = b*(1-a2) + cl.cshifts[j].destcolor[2]*a2;
    }

    v_blend[0] = r/255.0;
    v_blend[1] = g/255.0;
    v_blend[2] = b/255.0;
    v_blend[3] = a;
    if (v_blend[3] > 1)
        v_blend[3] = 1;
    if (v_blend[3] < 0)
        v_blend[3] = 0;
}

/*
=============
V_UpdatePalette
=============
*/
void V_UpdatePalette (void)
{
    int        i, j;
    qboolean    n;

    V_CalcPowerupCshift ();

    n = false;

    for (i=0 ; i<NUM_CSHIFTS ; i++)
    {
        if (cl.cshifts[i].percent != cl.prev_cshifts[i].percent)
        {
            n = true;
            cl.prev_cshifts[i].percent = cl.cshifts[i].percent;
        }
        for (j=0 ; j<3 ; j++)
            if (cl.cshifts[i].destcolor[j] != cl.prev_cshifts[i].destcolor[j])
            {
                n = true;
                cl.prev_cshifts[i].destcolor[j] = cl.cshifts[i].destcolor[j];
            }
    }

// drop the damage value
    cl.cshifts[CSHIFT_DAMAGE].percent -= host_frametime*150;
    if (cl.cshifts[CSHIFT_DAMAGE].percent <= 0)
        cl.cshifts[CSHIFT_DAMAGE].percent = 0;

// drop the bonus value
    cl.cshifts[CSHIFT_BONUS].percent -= host_frametime*100;
    if (cl.cshifts[CSHIFT_BONUS].percent <= 0)
        cl.cshifts[CSHIFT_BONUS].percent = 0;

    if (!n)
        return;

    V_CalcBlend ();
}

/*
==============================================================================

                        VIEW RENDERING

==============================================================================
*/

float angledelta (float a)
{
    a = anglemod(a);
    if (a > 180)
        a -= 360;
    return a;
}

/*
==================
CalcGunAngle
==================
*/
void CalcGunAngle (void)
{
    float    yaw, pitch, move;
    static float oldyaw = 0;
    static float oldpitch = 0;

    yaw = r_refdef.viewangles[YAW];
    pitch = -r_refdef.viewangles[PITCH];

    yaw = angledelta(yaw - r_refdef.viewangles[YAW]) * 0.4;
    if (yaw > 10)
        yaw = 10;
    if (yaw < -10)
        yaw = -10;
    pitch = angledelta(-pitch - r_refdef.viewangles[PITCH]) * 0.4;
    if (pitch > 10)
        pitch = 10;
    if (pitch < -10)
        pitch = -10;
    move = host_frametime*20;
    if (yaw > oldyaw)
    {
        if (oldyaw + move < yaw)
            yaw = oldyaw + move;
    }
    else
    {
        if (oldyaw - move > yaw)
            yaw = oldyaw - move;
    }

    if (pitch > oldpitch)
    {
        if (oldpitch + move < pitch)
            pitch = oldpitch + move;
    }
    else
    {
        if (oldpitch - move > pitch)
            pitch = oldpitch - move;
    }

    oldyaw = yaw;
    oldpitch = pitch;

    cl.viewent.angles[YAW] = r_refdef.viewangles[YAW] + yaw;
    cl.viewent.angles[PITCH] = - (r_refdef.viewangles[PITCH] + pitch);

    float scale = Cvar_VariableValue("v_idlescale");
    cl.viewent.angles[ROLL] -= scale * 
        sin(cl.time * Cvar_VariableValue("v_iroll_cycle")) *
        Cvar_VariableValue("v_iroll_level");
    cl.viewent.angles[PITCH] -= scale * 
        sin(cl.time * Cvar_VariableValue("v_ipitch_cycle")) *
        Cvar_VariableValue("v_ipitch_level");
    cl.viewent.angles[YAW] -= scale * 
        sin(cl.time * Cvar_VariableValue("v_iyaw_cycle")) *
        Cvar_VariableValue("v_iyaw_level");
}

/*
==============
V_BoundOffsets
==============
*/
void V_BoundOffsets (void)
{
    entity_t    *ent;

    ent = &cl_entities[cl.viewentity];

// absolutely bound refresh reletive to entity clipping hull
// so the view can never be inside a solid wall

    if (r_refdef.origin[0] < ent->origin[0] - 14)
        r_refdef.origin[0] = ent->origin[0] - 14;
    else if (r_refdef.origin[0] > ent->origin[0] + 14)
        r_refdef.origin[0] = ent->origin[0] + 14;
    if (r_refdef.origin[1] < ent->origin[1] - 14)
        r_refdef.origin[1] = ent->origin[1] - 14;
    else if (r_refdef.origin[1] > ent->origin[1] + 14)
        r_refdef.origin[1] = ent->origin[1] + 14;
    if (r_refdef.origin[2] < ent->origin[2] - 22)
        r_refdef.origin[2] = ent->origin[2] - 22;
    else if (r_refdef.origin[2] > ent->origin[2] + 30)
        r_refdef.origin[2] = ent->origin[2] + 30;
}

/*
==============
V_AddIdle

Idle swaying
==============
*/
void V_AddIdle (void)
{
    float scale = Cvar_VariableValue("v_idlescale");
    r_refdef.viewangles[ROLL] += scale *
        sin(cl.time * Cvar_VariableValue("v_iroll_cycle")) *
        Cvar_VariableValue("v_iroll_level");
    r_refdef.viewangles[PITCH] += scale *
        sin(cl.time * Cvar_VariableValue("v_ipitch_cycle")) *
        Cvar_VariableValue("v_ipitch_level");
    r_refdef.viewangles[YAW] += scale *
        sin(cl.time * Cvar_VariableValue("v_iyaw_cycle")) *
        Cvar_VariableValue("v_iyaw_level");
}


/*
==============
V_CalcViewRoll

Roll is induced by movement and damage
==============
*/
void V_CalcViewRoll (void)
{
    float        side;

    side = V_CalcRoll (cl_entities[cl.viewentity].angles, cl.velocity);
    r_refdef.viewangles[ROLL] += side;

    if (v_dmg_time > 0)
    {
        r_refdef.viewangles[ROLL] += v_dmg_time /
            Cvar_VariableValue("v_kicktime") * v_dmg_roll;
        r_refdef.viewangles[PITCH] += v_dmg_time /
            Cvar_VariableValue("v_kicktime") * v_dmg_pitch;
        v_dmg_time -= host_frametime;
    }

    if (cl.stats[STAT_HEALTH] <= 0)
    {
        r_refdef.viewangles[ROLL] = 80;    // dead view angle
        return;
    }

}


/*
==================
V_CalcIntermissionRefdef

==================
*/
void V_CalcIntermissionRefdef (void)
{
    entity_t    *ent, *view;
    float        old;

    // ent is the player model (visible when out of body)
    ent = &cl_entities[cl.viewentity];
    // view is the weapon model (only visible from inside body)
    view = &cl.viewent;

    VectorCopy (ent->origin, r_refdef.origin);
    VectorCopy (ent->angles, r_refdef.viewangles);
    view->model = NULL;

    // always idle in intermission
    old = Cvar_VariableValue("v_idlescale");
    Cvar_SetValue("v_idlescale", 1);
    V_AddIdle ();
    Cvar_SetValue("v_idlescale", old);
}

/*
==================
V_CalcRefdef

==================
*/
void V_CalcRefdef (void)
{
    entity_t    *ent, *view;
    int            i;
    vec3_t        forward, right, up;
    vec3_t        angles;
    float        bob;
    static float oldz = 0;

    V_DriftPitch ();

// ent is the player model (visible when out of body)
    ent = &cl_entities[cl.viewentity];
// view is the weapon model (only visible from inside body)
    view = &cl.viewent;


// transform the view offset by the model's matrix to get the offset from
// model origin for the view
    ent->angles[YAW] = cl.viewangles[YAW];    // the model should face
                                        // the view dir
    ent->angles[PITCH] = -cl.viewangles[PITCH];    // the model should face
                                        // the view dir


    bob = V_CalcBob ();

// refresh position
    VectorCopy (ent->origin, r_refdef.origin);
    r_refdef.origin[2] += cl.viewheight + bob;

// never let it sit exactly on a node line, because a water plane can
// dissapear when viewed with the eye exactly on it.
// the server protocol only specifies to 1/16 pixel, so add 1/32 in each axis
    r_refdef.origin[0] += 1.0/32;
    r_refdef.origin[1] += 1.0/32;
    r_refdef.origin[2] += 1.0/32;

    VectorCopy (cl.viewangles, r_refdef.viewangles);
    V_CalcViewRoll ();
    V_AddIdle ();

// offsets
    angles[PITCH] = -ent->angles[PITCH];    // because entity pitches are
                                            //  actually backward
    angles[YAW] = ent->angles[YAW];
    angles[ROLL] = ent->angles[ROLL];

    AngleVectors (angles, forward, right, up);

    for (i=0 ; i<3 ; i++)
    {
        r_refdef.origin[i] += 
            Cvar_VariableValue("scr_ofsx") * forward[i] +
            Cvar_VariableValue("scr_ofsy") * right[i] +
            Cvar_VariableValue("scr_ofsz") * up[i];
    }

    V_BoundOffsets ();

// set up gun position
    VectorCopy (cl.viewangles, view->angles);

    CalcGunAngle ();

    VectorCopy (ent->origin, view->origin);
    view->origin[2] += cl.viewheight;

    for (i=0 ; i<3 ; i++)
        view->origin[i] += forward[i]*bob*0.4;
    view->origin[2] += bob + 2;

    view->model = cl.model_precache[cl.stats[STAT_WEAPON]];
    view->frame = cl.stats[STAT_WEAPONFRAME];
    view->recolor = false;
    view->alphaval = 1.0f;

// set up the refresh position
    VectorAdd (r_refdef.viewangles, cl.punchangle, r_refdef.viewangles);

    // smooth out stair step ups
    if (cl.onground && ent->origin[2] - oldz > 0)
    {
        float steptime;

        steptime = cl.time - cl.oldtime;
        if (steptime < 0)
            steptime = 0;

        oldz += steptime * 80;
        if (oldz > ent->origin[2])
            oldz = ent->origin[2];
        if (ent->origin[2] - oldz > 12)
            oldz = ent->origin[2] - 12;
        r_refdef.origin[2] += oldz - ent->origin[2];
        view->origin[2] += oldz - ent->origin[2];
    }
    else
        oldz = ent->origin[2];

    if (Cvar_VariableValue("chase_active"))
        Chase_Update ();
}

/*
==================
V_RenderView

The player's clipping box goes from (-16 -16 -24) to (16 16 32) from
the entity origin, so any view position inside that will be valid
==================
*/
void V_RenderView (void)
{
    if (con_forcedup)
        return;

// don't allow cheats in multiplayer
    if (cl.maxclients > 1)
    {
        Cvar_Set ("scr_ofsx", "0");
        Cvar_Set ("scr_ofsy", "0");
        Cvar_Set ("scr_ofsz", "0");
    }

    if (cl.intermission)
    {    // intermission / finale rendering
        V_CalcIntermissionRefdef ();
    }
    else
    {
        if (!cl.paused /* && (sv.maxclients > 1 || key_dest == key_game) */ )
            V_CalcRefdef ();
    }
    
    // We've calculated the refdef for this frame, fire the pre-render event
    service::FramePreRender(r_refdef);

    R_PushDlights ();

    R_RenderView ();
    
    // Fire the post-render event
    service::FramePostRender();
}

//============================================================================

/*
=============
V_Init
=============
*/
void V_Init (void)
{
    Cmd_AddCommand ("v_cshift", V_cshift_f);
    Cmd_AddCommand ("bf", V_BonusFlash_f);
    Cmd_AddCommand ("centerview", V_StartPitchDrift);

    Cvar_RegisterVariable ("v_centermove", "0.15");
    Cvar_RegisterVariable ("v_centerspeed", "500");

    Cvar_RegisterVariable ("v_iyaw_cycle", "2");
    Cvar_RegisterVariable ("v_iroll_cycle", "0.5");
    Cvar_RegisterVariable ("v_ipitch_cycle", "1");
    Cvar_RegisterVariable ("v_iyaw_level", "0.3");
    Cvar_RegisterVariable ("v_iroll_level", "0.1");
    Cvar_RegisterVariable ("v_ipitch_level", "0.3");

    Cvar_RegisterVariable ("v_idlescale", "0");
    Cvar_RegisterVariable ("crosshair", "0", true);
    Cvar_RegisterVariable ("cl_crossx", "0");
    Cvar_RegisterVariable ("cl_crossy", "0");
    Cvar_RegisterVariable ("gl_cshiftpercent", "100");

    Cvar_RegisterVariable ("scr_ofsx", "0");
    Cvar_RegisterVariable ("scr_ofsy", "0");
    Cvar_RegisterVariable ("scr_ofsz", "0");
    Cvar_RegisterVariable ("cl_rollspeed", "200");
    Cvar_RegisterVariable ("cl_rollangle", "2.0");
    Cvar_RegisterVariable ("cl_bob", "0.02");
    Cvar_RegisterVariable ("cl_bobcycle", "0.6");
    Cvar_RegisterVariable ("cl_bobup", "0.5");

    Cvar_RegisterVariable ("v_kicktime", "0.5");
    Cvar_RegisterVariable ("v_kickroll", "0.6");
    Cvar_RegisterVariable ("v_kickpitch", "0.6");
}


